import os
import cv2
import torch
from model import VideoSaliencyModel, YUVVideoSaliencyModel
import argparse
from pathlib import Path
from os.path import join

from utils import *
from torchvision import transforms
import numpy as np

from ObtainGrid import QpGrid
from YUVFileLoader import YUVFileLoader

print(f"PyTorch version: {torch.__version__}")
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
print(device)
print(f"CUDA version: {torch.version.cuda}")

def read_config_file(file_path):
    config = {}
    with open(file_path, 'r') as file:
        for line in file:
            if line.startswith('#'):
                continue
            if ':' in line:
                key, value = line.split(':', 1)
                key = key.strip()
                value = value.split('#')[0].strip()  # Ignore comments after the value
                try:
                    value = int(value)
                except ValueError:
                    pass
                config[key] = value
    return config

def create_path_if_not_exists(path):
    if not os.path.exists(path):
        os.makedirs(path)
        print(f"Path '{path}' created successfully.")
    else:
        print(f"Path '{path}' already exists.")

def create_model(is_yyy, args):
    model_type = None
    if is_yyy <= 2 or is_yyy == 4:
        model_type = VideoSaliencyModel
    elif is_yyy == 3:
        model_type = YUVVideoSaliencyModel

    model = model_type(
        transformer_in_channel=args.transformer_in_channel, 
        nhead=args.nhead,
        use_upsample=bool(args.decoder_upsample),
        num_hier=args.num_hier,
        num_clips=args.clip_size   
    )

    return model

def validate(model, args, is_yyy, video, file_name='test'):
    path = join(args.save_path, file_name)
    rgb_images_path = join(path, 'rgb_images')
    qpFile_path = str(Path(args.output_folder) / (file.name + '.qp'))
    qpGridPath = join(path, 'qpGrid')
    create_path_if_not_exists(path)
    create_path_if_not_exists(rgb_images_path)
    create_path_if_not_exists(qpGridPath)

    img_size = (video[0].shape[1], video[0].shape[0])

    qpgrid = QpGrid(qpFile_path, qpGridPath, args.qp_target, img_size, len(video), verbose=args.verbose)

    len_temporal = args.clip_size

    model.eval()

    # process in a sliding window fashion
    if len(video) >= 2*len_temporal-1:

        snippet = []
        for i in range(len(video)):
            #print('Frames:', {i})
                        # Plot video[i] to check if it is correct
            torch_img = torch_transform(video[i], is_yyy=is_yyy)

            snippet.append(torch_img)
            
            if i >= len_temporal-1:
                clip = torch.FloatTensor(torch.stack(snippet, dim=0)).unsqueeze(0)
                clip = clip.permute((0,2,1,3,4))

                process(model, clip, img_size, qpgrid,file_name, i)

                if args.rgb_images == 1:
                    img_rgb_save(video[i], rgb_images_path, i)

                # process first (len_temporal-1) frames
                if i < 2*len_temporal-2:
                    if args.rgb_images == 1:
                        img_rgb_save(video[i-len_temporal+1], rgb_images_path, i-len_temporal+1)
                    process(model, torch.flip(clip, [2]), img_size, qpgrid ,file_name,i-len_temporal+1)

                del snippet[0]
        qpgrid.qp_to_file()
    else:
        print (' more frames are needed')

def img_rgb_save(img, rgb_images_path, i):
    rgb_img = cv2.cvtColor(img, cv2.COLOR_YUV2BGR)
    
    cv2.imwrite(f'{rgb_images_path}/imagen_{i}.png', rgb_img)

def torch_transform_rgb(img):
    img_transform = transforms.Compose([
            transforms.Resize((224, 384)),
            transforms.ToTensor(),
            transforms.Normalize(
                [0.485, 0.456, 0.406],
                [0.229, 0.224, 0.225]
            )
    ])
    img = img_transform(img)
    return img

def torch_transform(img, is_yyy=1):
    # If requested, create an image with YYY channel
    if is_yyy == 2:
        # Extract Y channel
        img_y = img[:,:,0]

        img_yyy = np.stack((img_y, img_y, img_y), axis=-1)
        img_yyy = Image.fromarray(img_yyy.astype('uint8'), 'RGB')

        img_yyy = img_yyy.resize((384, 224))  # Resize as per original function
        img_yyy = transforms.ToTensor()(img_yyy)
          # Reversed for compatibility with PIL size order
        return img_yyy
    elif is_yyy == 1:
        return torch_transform_rgb(img)
    elif is_yyy == 3 or is_yyy == 4:
        img_yuv = Image.fromarray(img)
        img_yuv = img_yuv.resize((384, 224))  # Resize as per original function
        img_yuv = transforms.ToTensor()(img_yuv)

        return img_yuv

def blur(img):
    k_size = 11
    bl = cv2.GaussianBlur(img,(k_size,k_size),0)
    return bl

def process(model, clip, img_size, qpgrid, dname, idx):
    with torch.no_grad():
        smap = model(clip.to(device)).cpu().data[0] # TODO: check if this is correct wht CPU
    
    smap = smap.numpy()
    smap = cv2.resize(smap, (img_size[0], img_size[1]))
    smap = blur(smap)

    smap = torch.from_numpy(smap).unsqueeze(0)

    
    processed_img = img_save(smap, join(args.save_path, dname, str(idx))+".png", normalize=True, save=(args.saliency_map==1))
    qpgrid.process_image(processed_img, idx)



if __name__ == '__main__':
    is_yyy = 4

    """if is_yyy == 1:
        save_path_yyy = ""
    elif is_yyy == 2:
        save_path_yyy = "_yyy"
    elif is_yyy == 3 or is_yyy == 4:
        save_path_yyy = "_yuv"""
    
    parser = argparse.ArgumentParser()
    parser.add_argument('--file_path', default='', type=str)
    parser.add_argument('--cfg_path', type=str)
    parser.add_argument('--cfg_folder', type=str)
    parser.add_argument('--output_folder',default='/home/tfm2/ViNetOriginal/ModelToQP/qpfiles', type=str)
    parser.add_argument('--file_weight',default='/home/tfm2/ViNetOriginal/ViNet/2024-04-05_10-35_vinet.pt',type=str)
    parser.add_argument('--qp_target', default=0, type=int)
    parser.add_argument('--nhead',default=4, type=int)
    parser.add_argument('--num_encoder_layers',default=3, type=int)
    parser.add_argument('--transformer_in_channel',default=32, type=int)
    parser.add_argument('--save_path',default="/media/beegfs/home/v582/SCRATCH/ModelToQPoutput", type=str)
    parser.add_argument('--start_idx',default=-1, type=int)
    parser.add_argument('--num_parts',default=4, type=int)
    parser.add_argument('--multi_frame',default=0, type=int)
    parser.add_argument('--decoder_upsample',default=1, type=int)
    parser.add_argument('--num_decoder_layers',default=-1, type=int)
    parser.add_argument('--num_hier',default=3, type=int)
    parser.add_argument('--clip_size',default=32, type=int)
    parser.add_argument('--verbose',default=0, type=int)
    parser.add_argument('--rgb_images',default=0, type=int)
    parser.add_argument('--saliency_map',default=0, type=int)
    
    args = parser.parse_args()
    print(args)

    # Check if the width and height are provided
    if (args.cfg_path is None or args.cfg_folder is None) and args.file_weight is None:
        print("Please provide the file path or file folder, file weight, width, and height.")
        # Print the missing arguments
        if args.cfg_path is None:
            print("Missing cfg path.")
        if args.cfg_folder is None:
            print("Missing cfg folder.")
        if args.file_weight is None:
            print("Missing file weight.")
        exit()
        

    if args.verbose >= 1:
        print("Creating and loading model...")
    model = create_model(is_yyy, args)

    model.load_state_dict(torch.load(args.file_weight))

    model = model.to(device)
    torch.backends.cudnn.benchmark = False
    if args.verbose >= 1:
        print("Model loaded successfully.")

    if args.cfg_folder:
        cfg_files = [f for f in Path(args.cfg_folder).iterdir() if f.suffix == '.cfg']
        # discard the first 8 videos
        for file in cfg_files:
            config = read_config_file(file)
            print(f"Processing file: {file}")    
            yuv_path = os.path.join(args.file_path,config.get('InputFile'))
            if not os.path.exists(yuv_path):
                print(f"File {config.get('InputFile')} does not exist. Skipping...")
                continue      
            video = YUVFileLoader.load_yuv_file(yuv_path, 
                                            int(config.get('SourceWidth')), 
                                            int(config.get('SourceHeight')), 
                                            is_10bit=config.get('InputBitDepth') == '10')
            create_path_if_not_exists(args.output_folder)
            validate(model, args, is_yyy, video, file.name)
            del video
    else:
        if args.cfg_path:
            config = read_config_file(args.cfg_path)
        else:
            config = {}
        video = YUVFileLoader.load_yuv_file(os.path.join(args.file_path,config.get('InputFile')), 
                                            int(config.get('SourceWidth')), 
                                            int(config.get('SourceHeight')), 
                                            is_10bit=config.get('InputBitDepth') == '10')
        file = Path(args.file_path)
        validate(model, args, is_yyy, video, file.name)

    print('Done')
